from abc import ABC, abstractmethod
from queue import SimpleQueue, Empty, Full

from eric_sse.message import MessageContract
from eric_sse.exception import NoMessagesException, RepositoryError


class Queue(ABC):
    """Abstract base class for queues (FIFO)"""
    @abstractmethod
    def pop(self) -> MessageContract:
        """
        Next message from the queue.

        Raises a :class:`eric_sse.exception.NoMessagesException` if the queue is empty.
        """
        ...

    @abstractmethod
    def push(self, message: MessageContract) -> None:
        ...

class InMemoryQueue(Queue):
    def __init__(self):
        self.__messages: list[MessageContract] = []
        self.__queue: SimpleQueue = SimpleQueue()

    def pop(self) -> MessageContract:
        try:
            m = self.__queue.get(block=False)
            return m
        except Empty:
            raise NoMessagesException

    def push(self, message: MessageContract) -> None:
        try:
            self.__queue.put(message)
        except Full as e:
            raise RepositoryError(e)






"""
This module is intended to those who want to create their own persistence layer.
A Redis implementation is available at https://pypi.org/project/eric-redis-queues/
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable, Any, AsyncIterable
from eric_sse.message import MessageContract
from eric_sse.listener import MessageQueueListener
from eric_sse.queue import Queue, InMemoryQueue

@dataclass
class Connection:
    """
    A connection is just a listener and its related message queue

    :param eric_sse.listener.MessageQueueListener listener:
    :param eric_sse.queue.Queue queue:
    """
    listener: MessageQueueListener
    queue: Queue

class ObjectPersistenceMixin(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        """Message type"""
        ...

    @property
    @abstractmethod
    def value_as_dict(self):
        ...

    @abstractmethod
    def setup_by_dict(self, setup: dict):
        ...

class ChannelInterface(ABC):

    @abstractmethod
    def open(self):
        ...

    @abstractmethod
    def add_listener(self) -> MessageQueueListener:
        ...

    @abstractmethod
    def register_connection(self, listener: MessageQueueListener, queue: Queue):
        ...

    @abstractmethod
    def remove_listener(self, listener_id: str):
        ...

    @abstractmethod
    def dispatch(self, listener_id: str, msg: MessageContract):
        ...

    @abstractmethod
    def broadcast(self, msg: MessageContract):
        ...

    @abstractmethod
    def get_listener(self, listener_id: str) -> MessageQueueListener:
        ...

    @abstractmethod
    def adapt(self, msg: MessageContract) -> Any:
        ...

    @abstractmethod
    async def message_stream(self, listener: MessageQueueListener) -> AsyncIterable[Any]:
        ...

class ChannelRepositoryInterface(ABC):

    @abstractmethod
    def load(self) -> Iterable[ChannelInterface]:
        ...

    @abstractmethod
    def persist(self, channel: ChannelInterface):
        ...

    @abstractmethod
    def delete(self, channel_id: str):
        ...


class ConnectionRepositoryInterface(ABC):
    """
    Abstraction for connections creation

    see :class:`eric_sse.entities.AbstractChannel`
    """

    @abstractmethod
    def create_queue(self, listener_id: str) -> Queue:
        """Returns a concrete :class:`eric_sse.connection.Connection`"""
        ...

    @abstractmethod
    def persist(self, connection: Connection) -> None:
        """Persists a concrete :class:`eric_sse.connection.Connection`"""
        ...

    @abstractmethod
    def load(self) -> Iterable[Connection]:
        """Returns an Iterable of all persisted connections :class:`eric_sse.connection.Connection`"""
        ...

    @abstractmethod
    def delete(self, listener_id: str) -> None:
        """Removes a persisted :class:`eric_sse.connection.Connection` given its correspondant listener id"""
        ...



class InMemoryConnectionRepository(ConnectionRepositoryInterface):
    """
    Default implementation used by :class:`eric_sse.entities.AbstractChannel`
    """
    def create_queue(self, listener_id: str) -> Queue:
        return InMemoryQueue()

    def persist(self, connection: Connection) -> None:
        pass

    def load(self) ->  Iterable[Connection]:
        return []

    def delete(self, listener_id: str) -> None:
        pass



from abc import ABC, abstractmethod

class MessageContract(ABC):
    """
    Contract class for messages

    A message is just a container of information identified by a type.
    For validation purposes you can override :class:`eric_sse.listener.MessageQueueListener.on_message`

    """
    @property
    @abstractmethod
    def type(self) -> str:
        """Message type"""
        ...

    @property
    @abstractmethod
    def payload(self) -> dict | list | str | int | float | None:
        """Message payload"""
        ...


class Message(MessageContract):
    """
    Models a simple message
    """
    def __init__(self, msg_type: str, msg_payload: dict | list | str | int | float | None = None) -> None:
        self._type = msg_type
        self._payload = msg_payload

    @property
    def type(self) -> str:
        return self._type

    @property
    def payload(self) -> dict | list | str | int | float | None:
        return self._payload


class UniqueMessage(MessageContract):
    """Messages plus an unique identifier"""
    __message: MessageContract
    __sender_id: str | None = None

    def __init__(self, message_id: str, message: MessageContract, sender_id: str = None) -> None:
        self.__id = message_id
        self.__message = message
        self.__sender_id = sender_id

    @property
    def id(self) -> str:
        """Unique message identifier"""
        return str(self.__id)

    @property
    def type(self) -> str:
        return  self.__message.type

    @property
    def sender_id(self) -> str:
        """Returns the id of the listener that sent the message"""
        return self.__sender_id

    @property
    def payload(self) -> dict | list | str | int | float | None:
        return self.__message.payload


class SignedMessage(Message):
    """Message plus sender id"""

    def __init__(self, sender_id: str, msg_type: str, msg_payload: dict | list | str | int | float | None = None):
        super().__init__(msg_type, msg_payload)
        self.__sender_id = sender_id

    @property
    def sender_id(self) -> str:
        """Returns the id of the listener that sent the message"""
        return self.__sender_id





import json
from pathlib import Path
from asyncio import open_unix_connection
from typing import AsyncIterable

class SocketClient:
    """
    A little facade to interact with SocketServer
    """

    def __init__(self, file_descriptor_path: str):
        self.__descriptor_path = Path(file_descriptor_path)

    async def send_payload(self, payload: dict):
        """
        Send an arbitrary payload to a socket

        see :class:`eric_sse.servers.SocketServer`
        """
        r, w = await open_unix_connection(self.__descriptor_path)
        w.write(json.dumps(payload).encode())
        w.write_eof()

        response = await r.read()
        await w.drain()
        w.close()
        await w.wait_closed()
        return response.decode()

    async def create_channel(self) -> str:
        return await self.send_payload({
            'v': 'c',
        })

    async def register(self, channel_id: str):
        return await self.send_payload({
            'v': 'r',
            'c': channel_id,
        })

    async def stream(self, channel_id, listener_id) -> AsyncIterable[str]:
        r, w = await open_unix_connection(self.__descriptor_path)
        payload = {
            'c': channel_id,
            'r': listener_id,
            'v': 'l',
        }
        w.write(json.dumps(payload).encode())
        w.write_eof()
        await w.drain()

        while not r.at_eof():
            m = await r.readline()
            yield m.decode()

    async def broadcast_message(self, channel_id: str, message_type: str, payload: str | dict | int | float):
        return await self.send_payload({
            'v': 'b',
            'c': channel_id,
            't': message_type,
            'p': payload
        })

    async def dispatch(self, channel_id: str, receiver_id: str, message_type: str, payload: str | dict | int | float):
        return await self.send_payload({
            'v': 'd',
            'c': channel_id,
            'r': receiver_id,
            't': message_type,
            'p': payload
        })

    async def remove_listener(self, channel_id: str, listener_id: str):
        return await self.send_payload({
            'v': 'rl',
            'c': channel_id,
            'r': listener_id
        })

    async def remove_channel(self, channel_id: str):
        return await self.send_payload({
            'v': 'rc',
            'c': channel_id
        })
